#include "wrapper.h"

namespace wrapper
{
  /* --------------------------------------------------------------------------------
   *  Private functions
   * -------------------------------------------------------------------------------- */
  //! Structures used in monitor result callbacks
  static struct CallbackMonitorContext
  {
    std::mutex mutex;             //!< Protect other member variables from being rewritten
    LE_CALLBACK_MONITOR callback; //!< Function object called by monitor data callback
    int format;                   //!< Format of monitor data
  } ctx_monitor_{};
  //! Structure used in progress callbacks
  static struct CallbackProgressContext
  {
    std::mutex mutex;              //!< Protect other member variables from being rewritten
    LE_CALLBACK_PROGRESS callback; //!< Function object called by progress callback
  } ctx_progress_{};

  static LE_MONDATAFORMAT_VARIANT makeMonDataVariant(void *ptr, int bytes, int format)
  {
    LE_MONDATAFORMAT_VARIANT v;
    switch (format)
    {
    case LE_MONDATA_FORMAT_1:
    {
      auto size = bytes / sizeof(LE_MONDATAFORMAT1);
      auto begin = reinterpret_cast<LE_MONDATAFORMAT1 *>(ptr);
      auto end = begin + size;
      v = std::vector<LE_MONDATAFORMAT1>(begin, end);
      break;
    }
    case LE_MONDATA_FORMAT_2:
    {
      auto size = bytes / sizeof(LE_MONDATAFORMAT2);
      auto begin = reinterpret_cast<LE_MONDATAFORMAT2 *>(ptr);
      auto end = begin + size;
      v = std::vector<LE_MONDATAFORMAT2>(begin, end);
      break;
    }
    default:
      // error
      break;
    }
    return v;
  }

  static void setCallbackMonitor(LE_CALLBACK_MONITOR &callback, int iFormat)
  {
    std::lock_guard<std::mutex> lock(ctx_monitor_.mutex);
    ctx_monitor_.callback = callback;
    switch (iFormat)
    {
    case LE_MONDATA_FORMAT_1:
    case LE_MONDATA_FORMAT_2:
      ctx_monitor_.format = iFormat;
      break;
    default:
      ctx_monitor_.format = LE_MONDATA_FORMAT_1;
      break;
    }
  }
  static void resetCallbackMonitor()
  {
    std::lock_guard<std::mutex> lock(ctx_monitor_.mutex);
    ctx_monitor_.callback = {};
  }
  static void callbackMonitor(void *ptr, int bytes)
  {
    std::lock_guard<std::mutex> lock(ctx_monitor_.mutex);
    auto data = makeMonDataVariant(ptr, bytes, ctx_monitor_.format);
    ctx_monitor_.callback(data);
  }
  static void setCallbackProgress(LE_CALLBACK_PROGRESS &callback)
  {
    std::lock_guard<std::mutex> lock(ctx_progress_.mutex);
    ctx_progress_.callback = callback;
  }
  static void resetCallbackProgress()
  {
    std::lock_guard<std::mutex> lock(ctx_progress_.mutex);
    ctx_progress_.callback = {};
  }
  static void callbackProgress(unsigned long progressitem, unsigned long value)
  {
    std::lock_guard<std::mutex> lock(ctx_progress_.mutex);
    ctx_progress_.callback(progressitem, value);
  }

  /* --------------------------------------------------------------------------------
   * Public (Wrapper) functions
   *
   * Return true if result of LE* is LE_SUCCESS, false otherwise
   * -------------------------------------------------------------------------------- */
#define LE_FAIL_THROW(expr)      \
  do                             \
  {                              \
    if (expr == LE_FAIL)         \
    {                            \
      throw LEPCError(__func__); \
    }                            \
  } while (0)

  short getDllVersion()
  {
    short version;
    LE_FAIL_THROW(LEGetDllVersion(&version));
    return version;
  }
  bool connect(std::string com)
  {
    return LEConnect(&com[0]) == LE_SUCCESS;
  }
  bool disconnect()
  {
    return LEDisconnect() == LE_SUCCESS;
  }
  LE_SYSTIME getDateTime()
  {
    auto time = LE_SYSTIME();
    LE_FAIL_THROW(LEGetDateTime(&time));
    return time;
  }
  bool setDateTime(LE_SYSTIME time)
  {
    return LESetDateTime(time) == LE_SUCCESS;
  }

  bool startOnlineCallBack(LE_CALLBACK_MONITOR &callback, int iFormat)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartOnlineCallBack(callbackMonitor, iFormat) == LE_SUCCESS;
  }
  bool startManualCallBack(LE_CALLBACK_MONITOR &callback, int iFormat, unsigned long ulDelayTime, unsigned long ulIdleTime, char fRepeat)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartManualCallBack(callbackMonitor, iFormat, ulDelayTime, ulIdleTime, fRepeat) == LE_SUCCESS;
  }
  bool startBufferCallBack(LE_CALLBACK_MONITOR &callback, int iFormat, unsigned long ulDelayTime, unsigned long ulIdleTime, char fRepeat, LE_BUFSIMDATA eData)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartBufferCallBack(callbackMonitor, iFormat, ulDelayTime, ulIdleTime, fRepeat, eData) == LE_SUCCESS;
  }
  bool startProgramCallBack(LE_CALLBACK_MONITOR &callback, int iFormat, LE_PROGRAM eProgram)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartProgramCallBack(callbackMonitor, iFormat, eProgram) == LE_SUCCESS;
  }
  bool startProgramCallBack2(LE_CALLBACK_MONITOR &callback, int iFormat, LE_PROGRAM eProgram, LE_CALLBACK_PROGRESS &callback_progress)
  {
    setCallbackMonitor(callback, iFormat);
    setCallbackProgress(callback_progress);
    return LEStartProgramCallBack2(callbackMonitor, iFormat, eProgram, callbackProgress) == LE_SUCCESS;
  }
  bool startFlowSendCharCallBack(LE_CALLBACK_MONITOR &callback, int iFormat, char fInitial, char cXon, char cXoff, LE_STDTABLENO eTable, unsigned long ulIdleTime)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartFlowSendCharCallBack(callbackMonitor, iFormat, fInitial, cXon, cXoff, eTable, ulIdleTime) == LE_SUCCESS;
  }
  bool startFlowSendLineCallBack(LE_CALLBACK_MONITOR &callback, int iFormat, char fInitial, LE_FLOWWATCH eWatch, LE_FLOWOPERATE eOperate, LE_STDTABLENO eTable, unsigned long ulIdleTime)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartFlowSendLineCallBack(callbackMonitor, iFormat, fInitial, eWatch, eOperate, eTable, ulIdleTime) == LE_SUCCESS;
  }
  bool startFlowRecvCharCallBack(LE_CALLBACK_MONITOR &callback, int iFormat, char fInitial, char cXon, char cXoff, unsigned long ulOnCount, unsigned long ulOffTime)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartFlowRecvCharCallBack(callbackMonitor, iFormat, fInitial, cXon, cXoff, ulOnCount, ulOffTime) == LE_SUCCESS;
  }
  bool startFlowRecvLineCallBack(LE_CALLBACK_MONITOR &callback, int iFormat, char fInitial, LE_FLOWWATCH eWatch, LE_FLOWOPERATE eOperate, unsigned long ulOnCount, unsigned long ulOffTime)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartFlowRecvLineCallBack(callbackMonitor, iFormat, fInitial, eWatch, eOperate, ulOnCount, ulOffTime) == LE_SUCCESS;
  }
  bool startEchoBufferCallBack(LE_CALLBACK_MONITOR &callback, int iFormat, unsigned long ulResponseTime)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartEchoBufferCallBack(callbackMonitor, iFormat, ulResponseTime) == LE_SUCCESS;
  }
  bool startEchoCharCallBack(LE_CALLBACK_MONITOR &callback, int iFormat)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartEchoCharCallBack(callbackMonitor, iFormat) == LE_SUCCESS;
  }
  bool startEchoLoopbackCallBack(LE_CALLBACK_MONITOR &callback, int iFormat, LE_LOOPBACKLINE eLoopbackLine)
  {
    setCallbackMonitor(callback, iFormat);
    return LEStartEchoLoopbackCallBack(callbackMonitor, iFormat, eLoopbackLine) == LE_SUCCESS;
  }

  bool startBertContinue(LE_BERTPATTERN ePattern, int flow)
  {
    return LEStartBertContinue(ePattern, flow) == LE_SUCCESS;
  }
  bool startBertRbit(LE_BERTPATTERN ePattern, LE_BERTCOUNT eCount, int flow)
  {
    return LEStartBertRbit(ePattern, eCount, flow) == LE_SUCCESS;
  }
  bool startBertRuntime(LE_BERTPATTERN ePattern, ULONG ulTestTime, int flow)
  {
    return LEStartBertRuntime(ePattern, ulTestTime, flow) == LE_SUCCESS;
  }
  bool startBertRepeat(LE_BERTPATTERN ePattern, short sTimeResolution, int flow)
  {
    return LEStartBertRepeat(ePattern, sTimeResolution, flow) == LE_SUCCESS;
  }

  bool restartCallBack(LE_CALLBACK_MONITOR &callback, int iFormat)
  {
    setCallbackMonitor(callback, iFormat);
    return LEReStartCallBack(callbackMonitor, iFormat) == LE_SUCCESS;
  }

  bool startOnline()
  {
    return LEStartOnline() == LE_SUCCESS;
  }
  bool stop()
  {
    if (LEStop() == LE_SUCCESS)
    {
      resetCallbackMonitor();
      resetCallbackProgress();
      return true;
    }
    return false;
  }
  int getRunStatus()
  {
    int state;
    if (LEGetRunStatus(&state) == LE_FAIL)
    {
      throw LEPCError(__func__);
    }
    return state;
  }
  bool emulationKey(int iCode)
  {
    return LEEmulationKey(iCode) == LE_SUCCESS;
  }
  LE_LTABLE getTable(int number)
  {
    auto table = LE_LTABLE();
    LE_FAIL_THROW(LEGetTable(number, &table));
    return table;
  }
  bool setTable(int number, LE_LTABLE table)
  {
    return LESetTable(number, table) == LE_SUCCESS;
  }
  bool setTableString(int number, std::string str)
  {
    return LESetTableString(number, &str[0]) == LE_SUCCESS;
  }
  LE_LPROGRAM getProgram(int number)
  {
    auto program = LE_LPROGRAM();
    LE_FAIL_THROW(LEGetProgram(number, &program));
    return program;
  }
  bool setProgram(int number, LE_LPROGRAM program)
  {
    return LESetProgram(number, program) == LE_SUCCESS;
  }

  bool setTrigger(int number, LE_TRIGGERFACTOR factor, LE_TRIGGERACTION action)
  {
    return LESetTrigger(number, factor, action) == LE_SUCCESS;
  }
  bool setTriggerOff(int number)
  {
    return LESetTriggerOff(number) == LE_SUCCESS;
  }
  std::tuple<char, LE_TRIGGERFACTOR, LE_TRIGGERACTION> getTrigger(int number)
  {
    char sw;
    auto factor = LE_TRIGGERFACTOR();
    auto action = LE_TRIGGERACTION();
    LE_FAIL_THROW(LEGetTrigger(number, &sw, &factor, &action));
    return std::make_tuple(sw, factor, action);
  }
  LE_LPROTOCOL getProtocol()
  {
    LE_LPROTOCOL eProtocol;
    LE_FAIL_THROW(LEGetProtocol(&eProtocol));
    return eProtocol;
  }
  int setProtocol(LE_LPROTOCOL eProtocol)
  {
    return LESetProtocol(eProtocol);
  }
  LE_SETUP_PROTOCOL getProtocolSetup()
  {
    auto protocol = LE_SETUP_PROTOCOL();
    LE_FAIL_THROW(LEGetProtocolSetup(&protocol));
    return protocol;
  }
  bool setProtocolSetup(LE_SETUP_PROTOCOL protocol)
  {
    return LESetProtocolSetup(protocol) == LE_SUCCESS;
  }

#ifdef __linux__
  ULONG getTimer(int number)
  {
    ULONG timer = 0;
    LE_FAIL_THROW(LEGetTimer(number, &timer));
    return timer;
  }
  ULONG getCounter(int number)
  {
    ULONG counter = 0;
    LE_FAIL_THROW(LEGetCounter(number, &counter));
    return counter;
  }
#endif

  bool interfaceRS232C(LE_IFMODE eMode, LE_IFPOLARITY ePolarity)
  {
    return LEInterfaceRS232C(eMode, ePolarity) == LE_SUCCESS;
  }
  bool interfaceRS530(LE_IFMODE eMode, LE_IFPOLARITY ePolarity, char fV35Select, LE_IFDRVCTRL eDrvCtrl)
  {
    return LEInterfaceRS530(eMode, ePolarity, fV35Select, eDrvCtrl) == LE_SUCCESS;
  }
  bool interfaceTTL(LE_IFLEVELEX eLevelEx, LE_IFPOLARITY ePolarity, LE_IFPOLARITY eClockPolarity, LE_IFOUTTYPE eOutType)
  {
    return LEInterfaceTTL(eLevelEx, ePolarity, eClockPolarity, eOutType) == LE_SUCCESS;
  }
  bool interfaceLineControlOff()
  {
    return LEInterfaceLineControlOff() == LE_SUCCESS;
  }
  bool interfaceLineControlDTE(unsigned short usDelayRsOn, unsigned short usDelaySd, unsigned short usDelayRsOff, char fErSw)
  {
    return LEInterfaceLineControlDTE(usDelayRsOn, usDelaySd, usDelayRsOff, fErSw) == LE_SUCCESS;
  }
  bool interfaceLineControlDCE(unsigned short usDelayCsOn, unsigned short usDelayCsOff, unsigned short usDelayCdOn, unsigned short usDelayRd, unsigned short usDelayCdOff)
  {
    return LEInterfaceLineControlDCE(usDelayCsOn, usDelayCsOff, usDelayCdOn, usDelayRd, usDelayCdOff) == LE_SUCCESS;
  }

  std::tuple<LE_IFPORT, LE_IFMODE, LE_IFPOLARITY, char, LE_IFDRVCTRL> getInterfaceMode()
  {
    LE_IFPORT ePort;
    LE_IFMODE eMode;
    LE_IFPOLARITY ePolarity;
    char fV35Select;
    LE_IFDRVCTRL eDrvCtrl;
    LE_FAIL_THROW(LEGetInterfaceMode(&ePort, &eMode, &ePolarity, &fV35Select, &eDrvCtrl));
    return {ePort, eMode, ePolarity, fV35Select, eDrvCtrl};
  }

  bool setIdleTimeRec(LE_IDLETIMEREC eIdleTimeRec)
  {
    return LESetIdleTimeRec(eIdleTimeRec) == LE_SUCCESS;
  }
  LE_IDLETIMEREC getIdleTimeRec()
  {
    LE_IDLETIMEREC eIdleTimeRec;
    LE_FAIL_THROW(LEGetIdleTimeRec(&eIdleTimeRec));
    return eIdleTimeRec;
  }
  bool setTimeStampRec(LE_TIMESTAMPREC eTimeStampRec)
  {
    return LESetTimeStampRec(eTimeStampRec) == LE_SUCCESS;
  }
  LE_TIMESTAMPREC getTimeStampRec()
  {
    LE_TIMESTAMPREC eTimeStampRec;
    LE_FAIL_THROW(LEGetTimeStampRec(&eTimeStampRec));
    return eTimeStampRec;
  }
  bool setLineRec(char fLineStateRec)
  {
    return LESetLineRec(fLineStateRec) == LE_SUCCESS;
  }
  bool getLineRec()
  {
    char fLineStateRec;
    LE_FAIL_THROW(LEGetLineRec(&fLineStateRec));
    return fLineStateRec;
  }
  bool setCaptureBuf(LE_CAPTUREBUF eCaptureBuf)
  {
    return LESetCaptureBuf(eCaptureBuf) == LE_SUCCESS;
  }
  LE_CAPTUREBUF getCaptureBuf()
  {
    LE_CAPTUREBUF eCaptureBuf;
    LE_FAIL_THROW(LEGetCaptureBuf(&eCaptureBuf));
    return eCaptureBuf;
  }

  LE_MONDATAFORMAT_VARIANT loadDT(std::string filename, int nFrames, int iFormat)
  {
    LE_MONDATAFORMAT_VARIANT v;
    switch (iFormat)
    {
    default:
    case LE_MONDATA_FORMAT_1:
    {
      auto frames = LE_MONDATAFORMAT1_FRAMES(nFrames);
      LE_FAIL_THROW(LELoadDT(&filename[0], frames.data(), &nFrames, LE_MONDATA_FORMAT_1));
      v = std::move(frames);
      break;
    }
    case LE_MONDATA_FORMAT_2:
    {
      auto frames = LE_MONDATAFORMAT2_FRAMES(nFrames);
      LE_FAIL_THROW(LELoadDT(&filename[0], frames.data(), &nFrames, LE_MONDATA_FORMAT_2));
      v = std::move(frames);
      break;
    }
    }
    return v;
  }

  bool isAbort()
  {
    return LEIsAbort() == LE_SUCCESS;
  }
  bool setBscTranslate(char fBscTranslate)
  {
    return LESetBscTranslate(fBscTranslate) == LE_SUCCESS;
  }
  char getBscTranslate()
  {
    char fBscTranslate;
    LE_FAIL_THROW(LEGetBscTranslate(&fBscTranslate));
    return fBscTranslate;
  }
  bool getScreen(std::string filename)
  {
    return LEGetScreen(&filename[0]) == LE_SUCCESS;
  }

  unsigned long getState()
  {
    unsigned long state;
    LE_FAIL_THROW(LEGetState(&state));
    return state;
  }
  int getBertDataCount()
  {
    int count;
    LE_FAIL_THROW(LEGetBertDataCount(&count));
    return count;
  }

  std::tuple<LE_DATABERT, int> getBertData()
  {
    auto bertdata = LE_DATABERT();
    int bertdatacount;
    LE_FAIL_THROW(LEGetBertData(&bertdata, &bertdatacount));
    return std::make_tuple(bertdata, bertdatacount);
  }

  bool importPcap(std::string filename)
  {
    return LEImportPcap(&filename[0]) == LE_SUCCESS;
  }
#undef LE_ERROR_THROW
} // namespace wrapper